/***************************************************************************
 *   Copyright (C) 2009, 2010 by Stephane List                             *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * Record a stream from an Axis camera
 * audio : AAC
 * video : MPEG4
 * Output file ca be .mp4 .mov or .3gp
 *
 * Convert it to flv with the following command line :
 * ffmpeg -i test.mp4 -ar 22050 -b 5000000 test.flv
 */
#include "recorder.h"
#include <string.h>
#include <QtDebug>
#include <QString>
#include <QSettings>

#define STREAM_INDEX_VIDEO 0
#define STREAM_INDEX_AUDIO 1
#define RTP_HEADER_LEN 12

#define VIDEO_RATE 90000
//FIXME no more used
#define AUDIO_RATE 8000 /* ne pas toucher */

const AVRational rtp_base_video={1,VIDEO_RATE};

#include "neaacdec.h"

#ifndef FAAD2_VERSION
#warning "FAAD1_VERSION : please use faad2 !"
#endif

recorder::recorder(QObject *parent)
: QObject(parent),
	recording_(false),
	header_written_(false),
	oc(NULL),
	contextv(NULL),
	audio_st(NULL),
	video_st(NULL),
	ts_first_frame_video(0)
{
}

recorder::~recorder()
{
	stop();
}

// returns 0 if OK, and < 0 if error
int recorder::open_video(AVStream *st)
{
    AVCodec *codec;
    AVCodecContext *c;

    c = st->codec;
    codec = avcodec_find_encoder(c->codec_id);
    if (!codec)
    {
        qDebug() << "video codec not found";
        return -1;
    }
    if (avcodec_open(c, codec) < 0)
    {
        qDebug() << "could not open video codec";
        return -1;
    }
    return 0;
}

// returns 0 if OK, and < 0 if error
int recorder::open_audio(AVStream *st)
{
	AVCodecContext *c;
	AVCodec *codec;

	c = st->codec;
	codec = avcodec_find_encoder(c->codec_id);
	if (!codec)
	{
		qDebug() << "audio codec not found";
		return -1;
	}
	if (avcodec_open(c, codec) < 0)
	{
		qDebug() << "could not open audio codec";
		//return -1;
	}
	if (c->frame_size != 1024)
	{
		qDebug() << "frame_size (!= 1024) : " << c->frame_size;
	}
	av_set_pts_info(st, 33, 1, audio_bit_rate_);
	return 0;
}

// returns >=0 if OK, and < 0 if error
int recorder::get_size_from_mp4_frame(void *buf)
{
	const int PIC_SIZE=200000;
	AVCodecContext *contextdecodeur;
	AVCodec *decodeur;
	int ret = 0;
	int got_picture;
	AVFrame *picture = (AVFrame*)malloc(PIC_SIZE);

	decodeur = avcodec_find_decoder(CODEC_ID_MPEG4);
	if (decodeur == NULL)
	{
		qDebug() << "mp4_dec: codec not found, probably not enabled in ffmpeg library";
		return -1;
	}
	contextdecodeur = avcodec_alloc_context();
	if (avcodec_open(contextdecodeur, decodeur) < 0)
	{
		qDebug() << "mp4_dec: Error, could not open codec";
		return -1;
	}
	ret = avcodec_decode_video(contextdecodeur, picture, &got_picture, (uint8_t*)buf, PIC_SIZE);
	free(picture);

	video_width = contextdecodeur->width;
	video_height = contextdecodeur->height;
	has_b_frames = contextdecodeur->has_b_frames;
	if (has_b_frames)
	{
		qDebug() << "Video has B frames !";
	}
	avcodec_close(contextdecodeur);
	av_free(contextdecodeur);
	qDebug() << "recorder::get_size_from_mp4_frame ended, picture : " << video_width << "*" << video_height;
	return ret;
}


// returns >=0 if OK, and < 0 if error
int recorder::add_video_stream(CodecID codec_id)
{
	qDebug() << "add_video_stream";

	video_st = av_new_stream(oc, STREAM_INDEX_VIDEO);
	if (!video_st)
	{
		qDebug("Could not alloc stream");
		return -1;
	}
	contextv = video_st->codec;
	contextv->codec_id = codec_id;
	contextv->codec_type = CODEC_TYPE_VIDEO;
	contextv->bit_rate = 250000;
	/* resolution must be a multiple of two */
	contextv->width = video_width;
	contextv->height = video_height;
	/* time base: this is the fundamental unit of time (in seconds) in terms
	   of which frame timestamps are represented. for fixed-fps content,
	   timebase should be 1/framerate and timestamp increments should be
	   identically 1. */
	contextv->time_base.den = video_fps_;
	contextv->time_base.num = 1;
	contextv->gop_size = 12; /* emit one intra frame every twelve frames at most */
	contextv->pix_fmt = PIX_FMT_YUV420P;
	contextv->has_b_frames = has_b_frames;
	if (contextv->codec_id == CODEC_ID_MPEG2VIDEO) {
		/* just for testing, we also add B frames */
		contextv->max_b_frames = 2;
	}
	if (contextv->codec_id == CODEC_ID_MPEG1VIDEO){
		/* Needed to avoid using macroblocks in which some coeffs overflow.
		   This does not happen with normal video, it just happens here as
		   the motion of the chroma plane does not match the luma plane. */
		contextv->mb_decision=2;
	}

	// some formats want stream headers to be separate
	if(!strcmp(oc->oformat->name, "mp4") || !strcmp(oc->oformat->name, "mov") || !strcmp(oc->oformat->name, "3gp"))
	{
		contextv->flags |= CODEC_FLAG_GLOBAL_HEADER;
	}
	else
	{
		qWarning() << "Warning, format name :" << oc->oformat->name << "(should be .mov .mp4 or .3gp)";
	}
	return 0;
}


/*
 * add an audio output stream
 */
int recorder::add_audio_stream(CodecID codec_id)
{
	AVCodecContext *contexta;
	qDebug() << "add_audio_stream";

	audio_st = av_new_stream(oc, STREAM_INDEX_AUDIO);
	if (!audio_st) {
		fprintf(stderr, "Could not alloc stream\n");
		return -1;
	}
	contexta = audio_st->codec;
	contexta->codec_id = codec_id;
	contexta->codec_type = CODEC_TYPE_AUDIO;

	/* put sample parameters */
	contexta->bit_rate = 64000;
	contexta->sample_rate = audio_bit_rate_;
	contexta->channels = audio_channels_;
	contexta->flags |= CODEC_FLAG_GLOBAL_HEADER;
	return 0;
}


void recorder::start(const QString &file)
{
	filename_ = file;
	qDebug() << "Recording in file : " << filename_;

	/* initialize libavcodec, and register all codecs and formats */
	av_register_all();

	qDebug() << "Using libavcodec Version : " << LIBAVCODEC_IDENT;

	/* allocate the output media context */
	oc = av_alloc_format_context();
	if (!oc) {
		emit error("Could not allocate format context");
		return;
	}
	fmt = guess_format(NULL, filename_.toAscii().constData(), NULL);
	if (!fmt) {
		emit error("Could not find suitable output format");
		return;
	}
	oc->oformat = fmt;
	if (file.length() > (int)sizeof(oc->filename))
	{
		QString err;
		err = QString("filename too long (limit = ");
		err+= QString((int)sizeof(oc->filename));
		err+= QString(") : ");
		err+= QString(file.length());
		emit error(err);
		return;
	}
	strncpy(oc->filename, filename_.toAscii().constData(), sizeof(oc->filename));
	recording_ = true;
}

void recorder::pause() {}

void recorder::stop()
{
	if (recording_)
	{
		recording_=false;
		qDebug() << "recorder::stop";
		if (header_written_)
		{
			av_write_trailer(oc);
			url_fclose(oc->pb);
			header_written_ = false;
		}
		if (oc)
		{
			av_free(oc);
			oc = NULL;
		}
	}
}

int recorder::write_header()
{
	if (fmt->video_codec != CODEC_ID_NONE)
	{
		if (add_video_stream(fmt->video_codec) < 0)
		{
			qDebug() << "Could not add video stream";
			return -1;
		}
	}
	else
	{
		qDebug() << "NO VIDEO !";
		return -1;
	}
	Q_CHECK_PTR(video_st);
	Q_CHECK_PTR(contextv);
	contextv->time_base.den = video_fps_;
	contextv->time_base.num = 1;

	if (fmt->audio_codec != CODEC_ID_NONE)
	{
		if (add_audio_stream(fmt->audio_codec) < 0)
		{
			qDebug() << "Could not add audio stream";
			return -1;
		}
	}
	else
	{
		qDebug() << "NO AUDIO !";
		return -1;
	}
	Q_CHECK_PTR(audio_st);

	/* set the output parameters (must be done even if no parameters). */
	if (av_set_parameters(oc, NULL) < 0)
	{
		qDebug() << "Invalid output format parameters";
		return -1;
	}
	dump_format(oc, 0, filename_.toAscii().constData(), 1);

	/* now that all the parameters are set, we can open the audio and
	   video codecs and allocate the necessary encode buffers */
	if (video_st)
	{
		open_video(video_st);
	}
	if (audio_st)
	{
		open_audio(audio_st);
	}
	if (url_fopen(&oc->pb, filename_.toAscii().constData(), URL_WRONLY) < 0) {
		qDebug() << "Could not open " << filename_;
		return -1;
	}
	av_write_header(oc);
	return 0;
}

void recorder::on_VideoAvailable(const QByteArray& data)
{
	int ret = 0;
	AVPacket pkt;
	qint64 ts;

	if (!header_written_)
	{
		// decode the first frame to get the frame size
		ts  = (((quint8)*(data.constData()+4)) << 24);
		ts += (((quint8)*(data.constData()+5)) << 16);
		ts += (((quint8)*(data.constData()+6)) << 8);
		ts += (((quint8)*(data.constData()+7)) << 0);

		ts_first_frame_video = ts;
		ts = 0;

		if (!fmt)
		{
			emit error("MP4 format not known by ffmpeg library");
			return;
		}

		if (fmt->video_codec != CODEC_ID_NONE)
		{
			if ((get_size_from_mp4_frame(RTP_HEADER_LEN + (uint8_t *)data.constData()) < 0) || (video_width == 0))
			{
#if 1
				qDebug() << "Could not get video size, wait next frame..";
				return;
#else
				qDebug() << "Could not get video size, trying default value";
				video_width = 320;
				video_height = 240;
#endif
			}
			if (write_header() == 0)
			{
				av_init_packet(&pkt);
				pkt.pts = 0;
				pkt.flags |= PKT_FLAG_KEY;
				pkt.stream_index = STREAM_INDEX_VIDEO;
				pkt.data= RTP_HEADER_LEN + (uint8_t *)data.constData();
				pkt.size= data.size() - RTP_HEADER_LEN;
				ret = av_write_frame(oc, &pkt);
				header_written_ = true;
			}
			else
			{
				qDebug() << "Could not write header";
				ret = -1;
			}
		}
		else
		{
			qDebug() << "NO VIDEO associated to file format!";
			ret = -1;
		}
	}
	else
	{
		ts  = (((quint8)*(data.constData()+4)) << 24);
		ts += (((quint8)*(data.constData()+5)) << 16);
		ts += (((quint8)*(data.constData()+6)) << 8);
		ts += (((quint8)*(data.constData()+7)) << 0);
		ts -= ts_first_frame_video;

		av_init_packet(&pkt);
		pkt.pts = av_rescale_q(ts, rtp_base_video, video_frame_rate_);
		pkt.flags |= PKT_FLAG_KEY;
		pkt.stream_index = STREAM_INDEX_VIDEO;
		pkt.data= RTP_HEADER_LEN + (uint8_t *)data.constData();
		pkt.size= data.size() - RTP_HEADER_LEN;
		ret = av_write_frame(oc, &pkt);
	}
}

void recorder::on_AudioAvailable(const QByteArray &data)
{
	int ret = 0;
	AVPacket pkt;

	if (header_written_ && !data.isEmpty())
	{
		av_init_packet(&pkt);
		pkt.flags |= PKT_FLAG_KEY;
		pkt.stream_index = STREAM_INDEX_AUDIO;
		pkt.pts = 0;
		pkt.data = RTP_HEADER_LEN + audio_header_size_ + (uint8_t *)data.constData();
		pkt.size = data.size() - RTP_HEADER_LEN - audio_header_size_;
		ret = av_write_frame(oc, &pkt);
	}
}

int recorder::set_video_format(QString & f)
{
	video_format_ = f;
	return 0;
}

int recorder::set_video_fps(quint32 fps)
{
	video_fps_ = fps;

	video_frame_rate_.den = fps;
	video_frame_rate_.num = 1;
	return 0;
}

int recorder::set_video_bit_rate(quint32 b)
{
	video_bit_rate_ = b;
	return 0;
}

int recorder::set_audio_bit_rate(quint32 b)
{
	audio_bit_rate_ = b;
	return 0;
}

int recorder::set_audio_format(QString & f)
{
	audio_format_ = f;
	if (audio_format_ == "AAC")
		audio_header_size_ = 4;
	else
		audio_header_size_ = 0;
	return 0;
}

int recorder::set_audio_channels(quint32 c)
{
	audio_channels_ = c;
	return 0;
}
