/***************************************************************************
 *   Copyright (C) 2008, 2009,2010 by Stephane List                        *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#if ! defined(RTSPCLIENTUDP_H_20080506_102456_INCLUDED_)
#define RTSPCLIENTUDP_H_20080506_102456_INCLUDED_

#include <QString>
#include <QStringList>
#include <QtDebug>
#include <QTcpSocket>
#include <QUdpSocket>
#include <QTime>
#include <QRegExp>
#include <QHttp>
#include <QBuffer>
#include <QTimer>

#include "sdp.h"
#include "rtp.h"

#define WATCHDOG_TIMEOUT 5

typedef struct StatRTP {
		quint16 last_seq; // Latest sequence number received
		quint16 max_seq; // Highest sequence number received
		quint16 cycle_seq; // Extended highest sequence number received
		quint32 lost; // number of sequences lost
		bool initialized;
} StatRTP;

//------------------------------------------------------------------------
/// RtspClientUdp : classe pour client RTSP/UDP
//------------------------------------------------------------------------
class RtspClientUdp : public QObject
{
	Q_OBJECT

	public:
		RtspClientUdp(QObject *parent = 0);
		~RtspClientUdp();

		// Setup credential before connecting if necessary
		void SetCredentials(const QString& Login, const QString& Password);

		// Connect to the RTSP server (OPTIONS, DESCRIBE, SETUP)
		void connectToHost(const QString & hostName, quint16 port=554);

		// Get available services
		QString & get_sdp();
		quint32 get_sdplen();

		// Get streams video info
		QString & get_video_format();
		quint32 get_video_fps();
		quint32 get_video_bit_rate();
		QString & get_video_control();

		// Get streams audio info
		quint32 get_audio_bit_rate();
		quint32 get_audio_channels();
		QString & get_audio_format();
		QString & get_audio_control();

		// Send PLAY to the RTSP server
		void play();

		void disconnect();

		enum ClientState
		{
			UnconnectedState,
			OptionsState,
			ConfigureCplState,
			DescribeState,
			DescribeCredentialState,
			ReceivingSDP,
			SetupVideo,
			SetupAudio,
			WaitPlay,
			PlayingVideo,
			PlayingAudio,
			Playing,
			ReceivingData,
			Stopping
		};

		enum ServerType
		{
			ServerUnknow,
			ServerAxis207,
			ServerAxis210,
			ServerAxis215,
			ServerAxisM1031W
		};

		enum MediaType
		{
			MediaUnknow,
			MediaAudio,
			MediaVideo
		};

		quint8 frac_lost_;

                QString video_format_; // "MP4V-ES", "MPEG4", "H264", "NONE"

                int sdp_mpeg4_profile_level_id_; // 245
                QString sdp_mpeg4_video_config_; // 000001b003000001b508000001000000012000c488800F50A041E1463F

                int sdp_h264_packetization_mode_; // 1
                int sdp_h264_profile_level_id_; // 420029
                QString sdp_h264_sprop_parameter_sets_; // "Z0IAKeNQFAe2AtwEBAaQeJEV,aM48gA=="

	signals:
		void connected(); // then you can call play function (not before)
		void disconnected();
		void AudioAvailable(const QByteArray &);
		void VideoAvailable(const QByteArray &);
		void stateChanged(ClientState clientstate);
		void error(const QString &);

	private:
		void reqOPTIONS();
		void reqDESCRIBE();
		void reqPLAY();
		void reqPLAYvideo();
		void reqPLAYaudio();
		void reqSTOP();
		void reqSETUPvideo();
		void reqSETUPaudio();
		void receiveDATA(QByteArray);
		void receiveSDP();
		void parseSDP();
		void parseSDP_video();
                void parseSDP_video_mpeg4();
                void parseSDP_video_h264();
		void parseSDP_audio();
		void decode_rtsp(QByteArray);
		void decode_rtcp(QByteArray, MediaType);
                void update_state(ClientState);

		QString Authentication() const;
		QString Header(const QString& Name_p, const QString& Value_p) const;

		// PTZ functionnality
		void CheckPtzCapabilities();

		QTimer watchdog_;
		quint8 watchdog_value_;
		quint16 watchdog_test_;

		bool PtzCapabilities_;
		QHttp HttpSock_;
		QBuffer Reply_to_check_ptz_;

		quint16 udp_server_video_port;
		quint16 udp_server_audio_port;
		quint16 udp_client_video_port, udp_client_audio_port;
		QString serverName_;
		quint16 serverPort_;
		QTcpSocket tcpSocket;
		QUdpSocket aSocket;
		QUdpSocket acSocket;
		QUdpSocket vSocket;
		QUdpSocket vcSocket;



		QString audio_format_;
		quint32 video_fps_;
		quint32 video_bit_rate_;
		quint32 audio_bit_rate_;
		quint32 audio_channels_;

		QByteArray abuf, vbuf;
		quint16 alast_seq, vlast_seq;
		quint32 local_aSSRC, local_vSSRC;
		quint32 remote_aSSRC, remote_vSSRC;
		quint16 max_aSeq, max_vSeq;
		quint16 ncycle_aSeq, ncycle_vSeq;
		quint32 alost, vlost, alost_last, vlost_last;

		quint32 aLSR, vLSR;
		QTime t_aLSR, t_vLSR;

		bool afirst, vfirst;
		ClientState state_;
		ServerType ServerType_;
		quint32 CSeq;
		bool hasSession;
		QString session;
		qint32  SDPlen;
		QString SDP_string, SDP_audio, SDP_video;
		QString acontrol_, vcontrol_;
		QString contentbase_;
		QString Login_;
		QString Password_;
		bool use_credentials_;
		int timeout_tcp_;

	public slots:
                void OnPtz(const QString&);

	private slots:
		void receiveAudio(); /* Audio data */
		void receivecAudio();/* Audio control */
		void receiveVideo(); /* Video data */
		void receivecVideo();/* Video control */
		void readBytes();
		void handleError(QAbstractSocket::SocketError);
		void on_tcp_connected();
		void on_tcp_disconnected();
		void CheckPtzCapabilitiesAnalyse(int, bool);
		void httpRequestFinished(int, bool);
		void watchdog();
};

#endif
