/***************************************************************************
 *   Copyright (C) 2009, 2010 by Stephane List                             *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "video_decoder.h"
#include <QtGui>

VideoDecoder::VideoDecoder(const QString & videocodec)
    : mp4_extradata_("")
    , Grabber_(0)
    , codec_(0)
    , CodecFrame_(NULL)
    , WindowFrame_(NULL)
    , count_(0)
    , img_convert_ctx_(NULL)
{
    qDebug() << "video_decoder: init libavcodec Version " << LIBAVCODEC_IDENT;
    avcodec_init();
    avcodec_register_all(); 
    qDebug() << "video_decoder: codec = " << videocodec;
    video_format_ = videocodec;

    if (videocodec == "MP4V-ES")
    {
        codec_ = avcodec_find_decoder(CODEC_ID_MPEG4);
        Grabber_ = avcodec_alloc_context();
        Grabber_->workaround_bugs = FF_BUG_AUTODETECT;
        Grabber_->error_concealment = FF_EC_DEBLOCK | FF_EC_GUESS_MVS;
#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(52,0,0)
        Grabber_->error_resilience = FF_ER_CAREFUL;
#endif
#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(52,20,0)
        Grabber_->strict_std_compliance = FF_COMPLIANCE_EXPERIMENTAL;
        Grabber_->error_recognition = FF_ER_CAREFUL;
        Grabber_->dsp_mask |= FF_MM_FORCE | FF_MM_MMX | FF_MM_SSE | FF_MM_SSE2;
        Grabber_->debug = 0;
#endif
    }
    else if (videocodec == "H264")
    {
        codec_ = avcodec_find_decoder(CODEC_ID_H264);
        Grabber_ = avcodec_alloc_context();

        pc_.cookie = MAGIC_COOKIE;
        pc_.profile_idc = 0x42;
        pc_.profile_iop = 0;
        pc_.level_idc = 29;
        pc_.packetization_mode = 1;

        Grabber_->extradata = (uint8_t*) &pc_;
        Grabber_->extradata_size = sizeof(pc_);
    }
    if (codec_ == 0)
    {
        qWarning("video_dec: codec not found, probably not enabled in ffmpeg library");
        return;
    }
    CodecFrame_ = avcodec_alloc_frame();
    WindowFrame_ = avcodec_alloc_frame();
    if ((CodecFrame_ == NULL) || (WindowFrame_ == NULL))
    {
        qFatal("video_dec: Error, could not alloc frames");
        return;
    }
    qDebug("video_dec: init succeeded");
}

bool VideoDecoder::init()
{
    if (avcodec_open(Grabber_, codec_) < 0)
    {
        qWarning("video_dec: Error, could not open codec");
    }

    if (video_format_ == "MP4V-ES")
    {
        int FinishedFlag;

        QByteArray a;
        a.append(mp4_extradata_);
        QByteArray b = QByteArray::fromHex(a);
        avcodec_decode_video(Grabber_,CodecFrame_,&FinishedFlag, (uint8_t*)b.data(), b.size());
        //int r = avcodec_decode_video(Grabber_,CodecFrame_,&FinishedFlag, (uint8_t*)b.data(), b.size());
        //if (r<0)
        //    return false;
        // TODO : [mpeg4 @ 0x1be7eb0]header damaged
        // TODO : correct this !
    }
    return true;
}

VideoDecoder::~VideoDecoder()
{
    qDebug() << "video_dec: close start\n";
    if (NULL != CodecFrame_)
    {
        av_free(CodecFrame_);
    }
    if (NULL != WindowFrame_)
    {
        av_free(WindowFrame_);
    }
    if (0 != Grabber_)
    {
        avcodec_close(Grabber_);
        av_free(Grabber_);
    }
    if (NULL != img_convert_ctx_)
    {
        sws_freeContext(img_convert_ctx_);
    }
    qDebug() << "video_dec: close end\n";
}

/////////////////////////////////////////////////////////////////////
/// Décodeur
/// i pBytes_p : Trame à décoder
/// i Size_p   : Nombre d'octets à décoder
/// i Format_p : Format de pixels souhaité 
/// o PicSize_p: Dimensions de l'image
/// o Video_p  : Trame décodée
/// retour : Succès/échec
/////////////////////////////////////////////////////////////////////
bool VideoDecoder::Decode(const char* pBytes_p
                          ,size_t Size_p
                          ,int Format_p
                          ,QSize & PicSize_p
                          ,QByteArray& Video_p)
{
    bool Decoded = false;
    bool Error = false;
    bool Finished = false;
    uint8_t* Begin = reinterpret_cast<uint8_t*>(const_cast<char*>(pBytes_p));
    const uint8_t* End = reinterpret_cast<const uint8_t*>(&pBytes_p[Size_p]);
    int w,h;



    //qDebug() << "Size to decode :" << Size_p;

    // if frames are not allocated, return a failure
    if ((NULL == CodecFrame_) || (NULL == WindowFrame_))
    {
        qWarning("CodecFrame or WindowFrame not allocated");
        return false;
    }

    // reinit frames
    avcodec_get_frame_defaults(CodecFrame_);
    avcodec_get_frame_defaults(WindowFrame_);

    while (! Finished && ! Error && (Begin < End))
    {
        int FinishedFlag = 0;
        int Dec = avcodec_decode_video(Grabber_,CodecFrame_,&FinishedFlag,Begin,static_cast<int>(End - Begin));
        Finished = 0 != FinishedFlag;
        if (Dec > 0)
        {
            Begin += Dec;
        }
        else
        {
            qWarning("Error in video decoding");
            Error = true;
            return false;
        }
    }
    w = Grabber_->width;
    h = Grabber_->height;
    //	if ((w < 176) || (w > 640) || (h < 144) | (h > 480))
    if ((w < 176) || (w > 704) || (h < 144) | (h > 576)) // PTZ
    {
        qDebug() << "Incorrect decoded image size :" << w << " * " << h << ". Ignore image";
        return false;
    }
    if (Finished && ! Error)
    {
        PicSize_p.setWidth(w);
        PicSize_p.setHeight(h);

        // Retaille tampon
        Video_p.resize(avpicture_get_size(Format_p,w,h));

        // Attache tampon à objet image
        avpicture_fill(reinterpret_cast<AVPicture *>(WindowFrame_)
                       ,reinterpret_cast<uint8_t*>(Video_p.data())
                       ,Format_p
                       ,w
                       ,h);
#ifdef img_convert
        img_convert(reinterpret_cast<AVPicture *>(WindowFrame_)
                    ,Format_p
                    ,reinterpret_cast<AVPicture *>(CodecFrame_)
                    ,Grabber_->pix_fmt
                    ,w
                    ,h);
#else
        img_convert_ctx_ = sws_getCachedContext(img_convert_ctx_,
                                                w, h, PIX_FMT_YUV420P, // src_pix_fmt,
                                                w, h, PIX_FMT_RGB24, // dst_pix_fmt,
                                                SWS_BICUBIC, NULL, NULL, NULL);
        if (img_convert_ctx_ == NULL)
        {
            qWarning() << "Cannot initialize the conversion context!";
            return false;
        }
        if (sws_scale(img_convert_ctx_,
                      CodecFrame_->data, /* src */
                      CodecFrame_->linesize, /* srcStride */
                      0, /* srcSliceY */
                      h, /* srcSliceH */
                      WindowFrame_->data, /* dst */
                      WindowFrame_->linesize) < 0)
        {
            qWarning() << "sws_scale failed";
            return false;
        }
#endif
        Decoded = true;
        ++count_;
    }
    return Decoded;
}

void VideoDecoder::set_bitrate(int bitrate)
{
    Grabber_->bit_rate = bitrate;
}

void VideoDecoder::mp4_set_global_headers(QString & extradata)
{
    mp4_extradata_ = extradata;
}

void VideoDecoder::mp4_set_profile_id(int profile_id)
{
    Grabber_->profile = profile_id; // TODO : verify
}

void VideoDecoder::h264_set_packetization_mode(int mode)
{
    pc_.packetization_mode = mode;
}

// 0x42 00 29
void VideoDecoder::h264_set_profile_level_id(int profile)
{
    QByteArray a;
    a.setNum(profile);
    QByteArray b = QByteArray::fromHex(a);

    pc_.cookie = MAGIC_COOKIE;
    pc_.profile_idc = b[0]; // 0x42;
    pc_.profile_iop = b[1];  // 0;
    pc_.level_idc   = b[2]; // 0x29;
    Grabber_->extradata = (uint8_t*) &pc_;
    Grabber_->extradata_size = sizeof(pc_);
}

// "Z0IAKeNQFAe2AtwEBAaQeJEV,aM48gA=="
// = sps, pps in (base64)
void VideoDecoder::h264_set_sprop_parameter_sets(QString & sprop)
{
    int i = sprop.indexOf(",");
    QByteArray ba;
    ba.append(sprop.left(i));
    qDebug() << "sps" << ba;
    QByteArray sps = QByteArray::fromBase64(ba);
    qDebug() << "sps=" << sps.data();
    qDebug() << "sps=" << sps.toHex();

    ba.clear();
    ba.append(sprop.mid(i+1));
    qDebug() << "pps" << ba;
    QByteArray pps = QByteArray::fromBase64(ba);
    qDebug() << "pps=" << pps.data();
    qDebug() << "pps=" << pps.toHex();

    //TODO : pass it to ffmpeg
};
